// 3DText.cpp : Defines the entry point for the DLL application.
//

#include "stdafx.h"

static HWND hWnd;			// OpenGL Window Handle
static HDC hDC;				// Device Context (needed to create RC)
static HGLRC hRC;			// Rendering Context
static int imageDim[2];		// Dimensions of Offline Image Buffer
static GLuint fontBase;		// Base for Font CallLists
static GLYPHMETRICSFLOAT fontFormat[256];	// Descriptors of Font Format

// Creates OpenGL window and rendering context of the specified size.
extern "C" BOOL PASCAL CreateRC(int width, int height)
{
  /////////////////////////////
  // 1.) Create Window

  if ((width <= 0) || (height <= 0))
  {
    // invalid window size
    return FALSE;
  }

  // create window; do not need message handler, so use "Static" class
  hWnd = CreateWindowEx(0, "Static",
    "OpenGL Offline Rendering Context", WS_POPUP, 0, 0,
    width, height, NULL, NULL, GetModuleHandle(NULL), NULL);
  if (!hWnd)
  {
    // couldn't create a window
    return FALSE;
  }

  // get windows device context
  hDC = GetDC(hWnd);
  if (!hDC)
  {
    // no device context
    return FALSE;
  }

  /////////////////////////////
  // 2.) Check Supported Pixel Format

  // set a default OpenGL pixel format descriptor, ask for 24bit RGB
  static PIXELFORMATDESCRIPTOR pfd =
  {
    sizeof(PIXELFORMATDESCRIPTOR), 1,
    PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
    PFD_TYPE_RGBA, 24,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 0, 0,
    PFD_MAIN_PLANE, 0, 0, 0, 0
  };

  // find a suitable pixel format
  GLuint pixelFormat = ChoosePixelFormat(hDC,&pfd);
  if (!pixelFormat)
  {
        // desired format not supported
    return FALSE;
  }


  /////////////////////////////
  // 3.) Set Pixel Format

  // set pixel format
  if(!SetPixelFormat(hDC, pixelFormat, &pfd))
  {
    // error setting pixel format
    return FALSE;
  }

  /////////////////////////////
  // 4.) Create Rendering Context

  // create OpenGL rendering context
  hRC = wglCreateContext(hDC);
  if (!hRC)
  {
    // error creating RC
    return FALSE;
  }

  // activate OpenGL RC
  if(!wglMakeCurrent(hDC,hRC))
  {
    // activation of RC failed
    return FALSE;
  }

  /////////////////////////////
  // 5.) OpenGL Setup

  // init viewport to window size
  glViewport(0, 0, width, height);

  // set perspective projection
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45.0f,(GLfloat)width/(GLfloat)height,0.1f, 100.0f);

  // set eye position and viewing direction
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(0, 0, -10, 0, 0, 0, 0, -1, 0);

  // depth buffer setup
  glClearDepth(1.0f);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LEQUAL);

  // shading setup
  glShadeModel(GL_SMOOTH);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHTING);
  glEnable(GL_COLOR_MATERIAL);

  /////////////////////////////
  // 6.) Store Dimensions of Image Buffer

  imageDim[0] = width;
  imageDim[1] = height;

  return TRUE;
}

// deletes OpenGL window and rendering context
extern "C" void PASCAL DeleteRC()
{
  if (hRC)
  {
    // release and delete rendering context
    wglMakeCurrent(NULL, NULL);
    wglDeleteContext(hRC);
    hRC = NULL;
  }

  if (hDC)
  {
    // release device context
    ReleaseDC(hWnd, hDC);
    hDC = NULL;
  }

  if (hWnd)
  {
    // destroy window
    DestroyWindow(hWnd);
    hWnd = NULL;
  }

  imageDim[0] = 0;
  imageDim[1] = 0;
}

// creates a OpenGL font
extern "C" BOOL PASCAL CreateOutlineFont(char *name, BOOL bold, BOOL italic)
{
  // activate OpenGL RC
  if(!wglMakeCurrent(hDC,hRC))
  {
    // activation of RC failed
    return FALSE;
  }

  // CallLists for 256 characters
  fontBase = glGenLists(256);

  HFONT font = CreateFont(-12, 0, 0, 0, bold ? FW_BOLD : FW_NORMAL,
    italic, 0, 0, ANSI_CHARSET,
    OUT_TT_PRECIS, CLIP_DEFAULT_PRECIS,
    ANTIALIASED_QUALITY, FF_DONTCARE | DEFAULT_PITCH, name);
  if (!font)
  {
    // error creating windows font
    return FALSE;
  }

  HGDIOBJ oldObject = SelectObject(hDC, font);
  if (!oldObject)
  {
    // error selecting font
    return FALSE;
  }

  // convert to 3D outline font (20% depth)
  if (!wglUseFontOutlines(hDC, 0, 255, fontBase, 0.0f, 0.2f, WGL_FONT_POLYGONS, fontFormat))
  {
    // creating font outlines failed
    return FALSE;
  }

  // unselecting font
  SelectObject(hDC, oldObject);

  return TRUE;
}

// delete a OpenGL font
extern "C" void PASCAL DeleteOutlineFont()
{
  glDeleteLists(fontBase, 256);
}

// renders 3D text
extern "C" BOOL PASCAL RenderText(char *text, int size, int color, int rotX, int rotY, int rotZ, void *buffer)
{
  double scale = size / 16.;

  // activate OpenGL RC
  if(!wglMakeCurrent(hDC,hRC))
  {
    // activation of RC failed
    return FALSE;
  }

  // set eye position and viewing direction
  glLoadIdentity();
  gluLookAt(0, 0, -10, 0, 0, 0, 0, -1, 0);

  // set the current content of the external buffer as background
  glDrawPixels(imageDim[0], imageDim[1], GL_RGB, GL_UNSIGNED_BYTE, buffer);
  glClear(GL_DEPTH_BUFFER_BIT);

  // set text color
  glColor3d((color & 0xff) / 255.0, ((color >> 8) & 0xff) / 255.0, ((color >> 16) & 0xff) / 255.0);

  // compute the width and find the max. height of text
  double width = 0;
  double height = 0;
  for (unsigned int i = 0; i < strlen(text); i++)
  {
    width += fontFormat[text[i]].gmfCellIncX;
    height = max(height, fontFormat[text[i]].gmfBlackBoxY);
  }

  // scale font size
  glScaled(scale, scale, scale);
  glTranslated(-width / 2.0, -height / 2.0, 0.0);

  // rotate, translate and scale text according to parameters
  glRotated(rotX, 1.0, 0.0, 0.0);
  glRotated(rotY, 0.0, 1.0, 0.0);
  glRotated(rotZ, 0.0, 0.0, 1.0);

  // draw text
  glPushAttrib(GL_LIST_BIT);
  glListBase(fontBase);
  glCallLists(strlen(text), GL_UNSIGNED_BYTE, text);
  glPopAttrib();

  // read rendered pixels back into the external buffer
  glReadPixels(0, 0, imageDim[0], imageDim[1], GL_RGB, GL_UNSIGNED_BYTE, buffer);

  return TRUE;
}


BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    return TRUE;
}

