using System;
using System.Runtime.InteropServices;
using System.Drawing;
using System.Drawing.Imaging;

namespace Photo_Editor_Application
{
	/// <summary>
	/// Wrapper class for 3D Text DLL. All methods of this class are
	/// static because the 3D Text DLLdoes not support multiple
	/// instances.
	/// </summary>
	/// <requirement>F:image_3dtext</requirement>
	/// <requirement>F:image_3dtext_color</requirement>
	/// <requirement>F:image_3dtext_font</requirement>
	/// <requirement>F:image_3dtext_rotate</requirement>
	/// <requirement>C:image_3dtext_singleline</requirement>
	/// <requirement>C:image_3dtext_opengl</requirement>
	public class OGL3DText
	{
		private const string DLLNAME = "3DText.dll";

		// extern DLL function declaration
		[DllImport(DLLNAME)]
		private static extern bool CreateRC(int width, int height);
		[DllImport(DLLNAME)]
		private static extern void DeleteRC();
		[DllImport(DLLNAME, CharSet=CharSet.Ansi)]
		private static extern bool CreateOutlineFont(string name, bool bold, bool italic);
		[DllImport(DLLNAME)]
		private static extern void DeleteOutlineFont();
		[DllImport(DLLNAME, CharSet=CharSet.Ansi)]
		private unsafe static extern void RenderText(string text, int size, int color, int rotX, int rotY, int rotZ, void *buffer);

		/// <summary>
		/// Indicates if the OpenGL font needs to be recreated.
		/// </summary>
		private static bool fontChanged = true;
		/// <summary>
		/// Indicates if the OpenGL rendering context needs to be recreated.
		/// This is the case if its size (not its position) changed.
		/// </summary>
		private static bool sizeChanged = true;

		// private members for set properties
		private static string text = "";
		private static int fontSize = 24;
		private static string fontName = "Arial";
		private static bool fontBold = false;
		private static bool fontItalic = false;
		private static int fontColor = System.Drawing.Color.White.ToArgb();
		private static System.Drawing.Bitmap image = null;
		private static System.Drawing.Rectangle placement;
		private static int rotX = 0;
		private static int rotY = 0;
		private static int rotZ = 0;

		/// <summary>
		/// Sets a text string to render.
		/// </summary>
		static public string Text
		{
			set 
			{
				text = value;
			}
		}

		/// <summary>
		/// Sets the font. Only font name, size, bold and italic style
		/// properties are considered from the Font class.
		/// </summary>
		static public System.Drawing.Font Font
		{
			set 
			{
				// font size is changed via scaling
				// no need to recreate the font
				fontSize = System.Convert.ToInt32(value.Size);

				// all other attributes require to recreate the font
				if (fontName != value.Name)
				{
					fontName = value.Name;
					fontChanged = true;
				}

				if (fontItalic != value.Italic)
				{
					fontItalic = value.Italic;
					fontChanged = true;
				}

				if (fontBold != value.Bold)
				{
					fontBold = value.Bold;
					fontChanged = true;
				}
			}
		}

		/// <summary>
		/// Sets the text color.
		/// </summary>
		static public System.Drawing.Color Color
		{
			set 
			{
				fontColor = value.ToArgb();
			}
		}

		/// <summary>
		/// Sets the background image.
		/// </summary>
		static public System.Drawing.Bitmap Image
		{
			set 
			{
				image = value;
			}
		}

		/// <summary>
		/// Defines a rectangle within the background image where the text
		/// is rendered.
		/// </summary>
		static public System.Drawing.Rectangle Placement
		{
			set 
			{
				if (placement.Size != value.Size)
				{
					sizeChanged = true;
				}
				placement = value;
			}
		}

		/// <summary>
		/// Sets the rotation along the x-axis.
		/// </summary>
		static public int RotationX
		{
			set 
			{
				rotX = value;
			}
		}

		/// <summary>
		/// Sets the rotation along the y-axis.
		/// </summary>
		static public int RotationY
		{
			set 
			{
				rotY = value;
			}
		}

		/// <summary>
		/// Sets the rotation along the z-axis.
		/// </summary>
		static public int RotationZ
		{
			set 
			{
				rotZ = value;
			}
		}

		/// <summary>
		/// Renders three-dimensional text. Internally creates the rendering
		/// context and font if necessary.
		/// </summary>
		/// 
		/// <returns>The method returns a new bitmap with the rendered text.</returns>
		static public unsafe System.Drawing.Bitmap Render()
		{
			// no background set
			if (image == null)
				return null;

			// limit placement to boundaries of background image
			if (placement.Right > image.Width)
			{
				placement.X = image.Width - placement.Width;
			}
			if (placement.Bottom > image.Height)
			{
				placement.Y = image.Height - placement.Height;
			}
			if ((placement.X < 0) | (placement.Y < 0) | (placement.Width <= 0) | (placement.Height <= 0))
			{
				return null;
			}

			// create new rendering context?
			if (sizeChanged)
			{
				DeleteRC();
				CreateRC(placement.Width, placement.Height);
			}

			// create new font?
			if (fontChanged)
			{
				DeleteOutlineFont();
				CreateOutlineFont(fontName, fontBold, fontItalic);
			}

			// clone region of interest (ROI) to get OpenGL rendering buffer
			Bitmap roi = image.Clone(placement, PixelFormat.Format24bppRgb);

			// lock memory and render 3D text into it
			BitmapData renderData = roi.LockBits(new Rectangle(0, 0, roi.Width, roi.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
			byte *renderBuffer = (byte*) renderData.Scan0.ToPointer();
			RenderText(text, fontSize, fontColor, rotX, rotY, rotZ, renderBuffer);

			// clone the background to get full size output buffer
			Bitmap output = image.Clone(new Rectangle(0, 0, image.Width, image.Height), PixelFormat.Format24bppRgb);

			// get pointer to output buffer and copy ROI into it
			BitmapData targetData = output.LockBits(placement, ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
			byte *targetBuffer = (byte*) targetData.Scan0.ToPointer();

			// copy the rendered buffer (sub region) into the final image
			for (int y = 0; y < (roi.Height); y++)
			{
				// copy one line, each pixel 3 bytes (RGB)
				for (int x = 0; x < (3 * roi.Width); x++)
				{
					targetBuffer[targetData.Stride * y + x] = renderBuffer[renderData.Stride * y + x];
				}
			}

			// unlock memory
			roi.UnlockBits(renderData);
			output.UnlockBits(targetData);

			// return a new Bitmap containing the final image
			return new Bitmap(output, new Size(output.Width, output.Height));
		}

		/// <summary>
		/// Deletes font and rendering context. Must be called to destroy the
		/// rendering context and OpenGL outline font after calling Render().
		/// </summary>
		static public void Close()
		{
			DeleteOutlineFont();
			DeleteRC();
		}

	}
}
